#!/usr/bin/env python3

'''
JiffyBox external inventory script
==================================
'''

try:
    import configparser
except:
    import ConfigParser as configparser

import argparse
import os
import re
import sys
from ansible.errors import AnsibleError as ae
from time import time

import requests
import six

try:
    import json
except ImportError:
    import simplejson as json


class JiffyBoxInventory(object):

    def _empty_inventory(self):
        return {"_meta": {"hostvars": {}}}

    def __init__(self):
        ''' Main execution path '''

        # Inventory grouped by instance IDs, tags, security groups, regions,
        # and availability zones
        self.inventory = self._empty_inventory()

        # Index of hostname (address) to instance ID
        self.index = {}

        # Read settings and parse CLI arguments
        self.parse_cli_args()
        self.read_settings()

        # Cache
        if self.args.refresh_cache:
            self.do_api_calls_update_cache()
        elif not self.is_cache_valid():
            self.do_api_calls_update_cache()

        # Data to print
        if self.args.set_host_groups:
            data_to_print = self.set_host_groups()
        elif self.args.host:
            data_to_print = self.get_host_info()
        elif self.args.list:
            # Display list of instances for inventory
            if self.inventory == self._empty_inventory():
                data_to_print = self.get_inventory_from_cache()
            else:
                data_to_print = self.json_format_dict(self.inventory, True)
        else:
            data_to_print = {}

        print(data_to_print)

    def parse_cli_args(self):
        ''' Command line argument processing '''

        parser = argparse.ArgumentParser(description='Produce an Ansible Inventory file based on JiffyBox')
        parser.add_argument('--company', default=os.environ.get('ANSIBLE_COMPANY', None),
                    help='Only useful if inventories from multiple companies are installed')
        parser.add_argument('--list', action='store_true', default=True,
                            help='List instances (default: True)')
        parser.add_argument('--host', action='store',
                            help='Get all the variables about a specific instance')
        parser.add_argument('--refresh-cache', action='store_true', default=os.environ.get('ANSIBLE_INVENTORY_REFRESH', False),
                            help='Force refresh of cache by making API requests to JiffyBox (default: False - use cache files)')
        parser.add_argument('--set-host-groups', action='store',
                            help='Set the inventory groups for the host(s)')
        self.args = parser.parse_args()

    def read_settings(self):
        ''' Reads the settings from the jiffybox.ini file '''
        config = configparser.ConfigParser()
        jiffybox_default_ini_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'jiffybox.ini')
        jiffybox_ini_path = os.path.expanduser(os.path.expandvars(os.environ.get('JIFFYBOX_INI_PATH', jiffybox_default_ini_path)))
        config.read(jiffybox_ini_path)

        if config.has_option('jiffybox', 'expand_csv_tags'):
            self.expand_csv_tags = config.getboolean('jiffybox', 'expand_csv_tags')
        else:
            self.expand_csv_tags = False

        # Replace dash or not in group names
        if config.has_option('jiffybox', 'replace_dash_in_groups'):
            self.replace_dash_in_groups = config.getboolean('jiffybox', 'replace_dash_in_groups')
        else:
            self.replace_dash_in_groups = True

        # Return all JiffyBox instances?
        if config.has_option('jiffybox', 'all_instances'):
            self.all_instances = config.getboolean('jiffybox', 'all_instances')
        else:
            self.all_instances = False

        # Instance states to be gathered in inventory. Default is 'running'.
        # Setting 'all_instances' to 'yes' overrides this option.
        jiffybox_valid_instance_states = [
            'READY',
            'CREATING',
            'UPDATING',
            'CHANGING PLAN',
            'STATUS_READY',
            'STATUS_CREATING',
            'STATUS_UPDATING'
        ]
        self.jiffybox_instance_states = []
        if self.all_instances:
            self.jiffybox_instance_states = jiffybox_valid_instance_states
        elif config.has_option('jiffybox', 'instance_states'):
            for instance_state in config.get('jiffybox', 'instance_states').split(','):
                instance_state = instance_state.strip()
                if instance_state not in jiffybox_valid_instance_states:
                    continue
                self.jiffybox_instance_states.append(instance_state)
        else:
            self.jiffybox_instance_states = ['READY']

        # Cache related
        cache_dir = os.path.expanduser(config.get('jiffybox', 'cache_path'))
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)
        suffix = ''
        if self.args.company:
            suffix = '-' + self.args.company
        self.cache_path_cache = cache_dir + "/ansible-jiffybox" + suffix + ".cache"
        self.cache_path_index = cache_dir + "/ansible-jiffybox" + suffix + ".index"
        self.cache_max_age = config.getint('jiffybox', 'cache_max_age')

        # Do we need to just include hosts that match a pattern?
        try:
            pattern_include = config.get('jiffybox', 'pattern_include')
            if pattern_include and len(pattern_include) > 0:
                self.pattern_include = re.compile(pattern_include)
            else:
                self.pattern_include = None
        except configparser.NoOptionError:
            self.pattern_include = None

        # Do we need to exclude hosts that match a pattern?
        try:
            pattern_exclude = config.get('jiffybox', 'pattern_exclude')
            if pattern_exclude and len(pattern_exclude) > 0:
                self.pattern_exclude = re.compile(pattern_exclude)
            else:
                self.pattern_exclude = None
        except configparser.NoOptionError:
            self.pattern_exclude = None

    def is_cache_valid(self):
        ''' Determines if the cache files have expired, or if it is still valid '''

        if os.path.isfile(self.cache_path_cache):
            mod_time = os.path.getmtime(self.cache_path_cache)
            current_time = time()
            if (mod_time + self.cache_max_age) > current_time:
                if os.path.isfile(self.cache_path_index):
                    return True

        return False

    def do_api_calls_update_cache(self):
        ''' Do API calls to each region, and save data in cache files '''

        self.get_instances()
        self.write_to_cache(self.inventory, self.cache_path_cache)
        self.write_to_cache(self.index, self.cache_path_index)

    def get_instances(self):
        ''' Makes a JiffyBox API call to the list of instances '''

        try:
            conn = self.connect()
            instances = conn.get_all_instances()
            for instance in instances:
                self.add_instance(instances[instance])

        except ae as e:
            self.fail_with_error(e.message, 'getting JiffyBox instances')

    def connect(self):
        connect_args = {}

        conn = JiffyBoxConnect(self.args.company)
        # connect_to_region will fail "silently" by returning None if the region name is wrong or not supported
        if conn is None:
            self.fail_with_error("Something went wrong during connection")
        return conn

    def fail_with_error(self, err_msg, err_operation=None):
        '''log an error to std err for ansible-playbook to consume and exit'''
        if err_operation:
            err_msg = 'ERROR: "{err_msg}", while: {err_operation}'.format(
                err_msg=err_msg, err_operation=err_operation)
        sys.stderr.write(err_msg)
        sys.exit(1)

    def get_inventory_from_cache(self):
        ''' Reads the inventory from the cache file and returns it as a JSON
        object '''

        cache = open(self.cache_path_cache, 'r')
        json_inventory = cache.read()
        return json_inventory

    def load_index_from_cache(self):
        ''' Reads the index from the cache file sets self.index '''

        cache = open(self.cache_path_index, 'r')
        json_index = cache.read()
        self.index = json.loads(json_index)

    def write_to_cache(self, data, filename):
        ''' Writes data in JSON format to a file '''

        json_data = self.json_format_dict(data, True)
        cache = open(filename, 'w')
        cache.write(json_data)
        cache.close()

    def to_safe(self, word):
        ''' Converts 'bad' characters in a string to underscores so they can be used as Ansible groups '''
        regex = "[^A-Za-z0-9\_"
        if not self.replace_dash_in_groups:
            regex += "\-"
        return re.sub(regex + "]", "_", word)

    def json_format_dict(self, data, pretty=False):
        ''' Converts a dict to a JSON object and dumps it as a formatted
        string '''

        if pretty:
            return json.dumps(data, sort_keys=True, indent=2)
        else:
            return json.dumps(data)

    def get_instance(self, region, instance_id):
        conn = self.connect()

        instances = conn.get_all_instances()
        for instance in instances:
            if int(instance) == instance_id:
                return instances[instance]

        raise ae('Given host does not exist')

    def add_instance(self, instance):
        ''' Adds an instance to the inventory and index, as long as it is
        addressable '''

        # Only return instances with desired instance states
        if 'status' not in instance or instance['status'] not in self.jiffybox_instance_states:
            return

        # Only return running instances
        if 'running' not in instance or not instance['running']:
            return

        # Select the best destination address
        dest = None
        if 'name' in instance:
            dest = instance['name']
        elif 'ips' in instance and 'public' in instance['ips']:
            dest = instance['ips']['public'][0]

        if not dest:
            # Skip instances we cannot address (e.g. private VPC subnet)
            return

        # if we only want to include hosts that match a pattern, skip those that don't
        if self.pattern_include and not self.pattern_include.match(dest):
            return

        # if we need to exclude hosts that match a pattern, skip those
        if self.pattern_exclude and self.pattern_exclude.match(dest):
            return

        # Add to index
        self.index[dest] = ['all', instance['id']]

        # Inventory: Group by tag keys
        if 'metadata' in instance and 'ansibleGroups' in instance['metadata']:
            v = instance['metadata']['ansibleGroups']
            if 'running' in instance and not instance['running']:
                v += ',offline'
        else:
            v = 'offline'
        if self.expand_csv_tags and v and ',' in v:
            values = []
            for x in v.split(','):
                values.append(x.strip())
        else:
            values = [v]
        for v in values:
            self.push(self.inventory, self.to_safe(v), dest)

        # Global Tag: tag all JiffyBox instances
        self.push(self.inventory, 'jiffybox', dest)

        self.inventory["_meta"]["hostvars"][dest] = self.get_host_info_dict_from_instance(instance)

    def push(self, my_dict, key, element):
        ''' Push an element onto an array that may not have been defined in
        the dict '''
        group_info = my_dict.setdefault(key, [])
        if isinstance(group_info, dict):
            host_list = group_info.setdefault('hosts', [])
            host_list.append(element)
        else:
            group_info.append(element)

    def get_host_info_dict_from_instance(self, instance):
        instance_vars = {}
        for key in instance:
            value = instance[key]
            key = self.to_safe('jiffybox_' + key)

            # Handle complex types
            if type(value) in [int, bool]:
                instance_vars[key] = value
            elif key == 'jiffybox_ips':
                instance_vars[key] = value
            elif isinstance(value, six.string_types):
                instance_vars[key] = value.strip()
            elif type(value) == type(None):
                instance_vars[key] = ''
            elif key == 'jiffybox_metadata':
                for k, v in value.items():
                    if self.expand_csv_tags and ',' in v:
                        values = []
                        for x in v.split(','):
                            values.append(x.strip())
                        v = values
                    key = self.to_safe('jiffybox_metadata_' + k)
                    instance_vars[key] = v
            else:
                pass

        instance_vars['static_ipv4'] = instance_vars['jiffybox_ips']['public'][0]
        return instance_vars

    def get_host(self, host):
        ''' Get variables about a specific host '''

        if len(self.index) == 0:
            # Need to load index from cache
            self.load_index_from_cache()

        if not self.args.host in self.index:
            # try updating the cache
            self.do_api_calls_update_cache()
            if not self.args.host in self.index:
                # host might not exist anymore
                return self.json_format_dict({}, True)

        (region, instance_id) = self.index[self.args.host]

        return self.get_instance(region, instance_id)

    def set_host_groups(self):
        ''' Add host or hosts to a group '''

        instance = self.get_host(self.args.host)
        connect = self.connect()
        connect.set_metadata(instance, 'ansibleGroups', self.args.set_host_groups)
        return 'ok'

    def get_host_info(self):
        ''' Get variables about a specific host '''

        instance = self.get_host(self.args.host)
        return self.json_format_dict(self.get_host_info_dict_from_instance(instance), True)


class JiffyBoxConnect(object):

    configPath = '/etc/jiffybox.cfg'

    def __init__(self, profile):
        self.profile = profile
        self._readConfig()
        self.url = 'https://api.jiffybox.de/' + self.api_token + '/v1.0/'

    def get_all_instances(self):
        return self._request('jiffyBoxes')

    def set_metadata(self, instance, key, value):
        self._request('jiffyBoxes/' + str(instance['id']), {'metadata[' + key + ']': value})

    def _readConfig(self):
        config = configparser.ConfigParser()
        config.read(self.configPath)

        if self.profile:
            section = 'profile ' + self.profile
        else:
            section = 'Credentials'
        if config.has_option(section, 'api_token'):
            self.api_token = config.get(section, 'api_token')
        else:
            raise ae('Can not find credentials')

    def _request(self, path, data = None, method = 'GET'):
        encoder = json.JSONEncoder()
        postData = {}

        if data:
            method = 'POST'
            for key in data:
                item = data.get(key)
                if type(item) is list or type(item) is dict:
                    if len(item) > 0:
                        item = encoder.encode(item)
                if type(item) is int or type(item) is unicode or type(item) is bool:
                    item = str(item)
                if item and type(item) is str and len(item) > 0:
                    postData.__setitem__(key, item)

        request_result = {}
        try:
            if method == 'GET':
                request_result = requests.get(self.url + path)
            elif method == 'POST':
                request_result = requests.put(self.url + path, data = postData)
            elif method == 'DELETE':
                request_result = requests.delete(self.url + path)
        except ae as e:
            raise ae('No result from JiffyBox API')

        decoder = json.JSONDecoder()
        content = decoder.decode(request_result.content.decode('utf8'))
        if not content['result']:
            msg = content['messages']
            raise ae('%s' % msg)

        return content['result']


# Run the script
JiffyBoxInventory()
