# python 3 headers, required if submitting to Ansible
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = """
        lookup: drupal_sites_from_git
        author: Juergen Haas <juergen.haas@lakedrops.com>
        version_added: "0.9"
        short_description: get drupal sites from git branches
        description:
            - This lookup returns definition for the variable drupal_settings derived from a git repository.
        options:
          _terms:
            description: repository URL(s)
            required: True
"""
from ansible.errors import AnsibleError, AnsibleParserError
from ansible.plugins.lookup import LookupBase
from subprocess import Popen, PIPE, STDOUT
from six import string_types
import os

try:
    from __main__ import display
except ImportError:
    from ansible.utils.display import Display
    display = Display()


class LookupModule(LookupBase):

    def run(self, terms, variables=None, drupal_project_name="", drupal_config_repos=[], drupal_domain_pattern="", drupal_domains_pattern="", **kwargs):
        configrepos = {}
        for configrepo in drupal_config_repos:
            display.debug("Config repository lookup using %s" % configrepo)
            display.vvvv(u"Config repository lookup using %s" % configrepo)
            configproject = os.path.splitext(os.path.basename(configrepo))[0]
            configrepos.__setitem__(configproject, {
                "repo": configrepo,
                "branches": []
            })
            try:
                cmd = [
                    'git',
                    'ls-remote',
                    '--heads',
                    configrepo
                ]
                process = Popen(cmd, stdout=PIPE, stderr=STDOUT)
                process.wait()
                for line in process.stdout.read().decode('utf8').split("\n"):
                    if not line:
                        continue
                    parts = line.split("\t")
                    if len(parts) != 2:
                        continue
                    fullbranch = parts[1]
                    configrepos[configproject]["branches"].append(fullbranch.replace('refs/heads/', ''))
            except AnsibleParserError:
                raise AnsibleError("could not access repository in lookup: %s" % configrepo)

        ret = []

        repomode = True
        lines = []
        if terms[0] == "":
            terms = [drupal_project_name]
            repomode = False
            dummyoutput = {}
            for configproject in configrepos:
                for branch in configrepos[configproject]["branches"]:
                    dummyoutput[branch] = drupal_project_name + "\t" + branch
            lines = []
            for line in dummyoutput:
                lines.append(dummyoutput[line])
        for term in terms:
            display.debug("Repository lookup using %s" % term)
            display.vvvv(u"Repository lookup using %s" % term)
            try:
                if repomode:
                    cmd = [
                        'git',
                        'ls-remote',
                        '--heads',
                        term
                    ]
                    process = Popen(cmd, stdout=PIPE, stderr=STDOUT)
                    process.wait()
                    lines = process.stdout.read().decode('utf8').split("\n")
                    project = os.path.splitext(os.path.basename(term))[0]
                else:
                    project = drupal_project_name
                for line in lines:
                    if not line:
                        continue
                    parts = line.split("\t")
                    if len(parts) != 2:
                        continue
                    fullbranch = parts[1]
                    branch = fullbranch.replace('refs/heads/', '')
                    sanitizedbranch = branch.replace('/', '-').replace('_', '-')
                    ismaster = 'true' if (branch == 'master') else 'false'
                    isnotmaster = 'false' if (branch == 'master') else 'true'

                    values = {
                        "repo": term,
                        "project": project,
                        "fullbranch": fullbranch,
                        "branch": branch,
                        "sanitizedbranch": sanitizedbranch,
                        "branch2url": sanitizedbranch,
                        "ismaster": ismaster,
                        "isnotmaster": isnotmaster
                    }
                    if branch == 'master':
                        values.__setitem__("branch2url", 'live')
                    elif branch == 'develop':
                        values.__setitem__("branch2url", 'test')
                    if len(configrepos) > 0:
                        for configproject in configrepos:
                            workvalues = values.copy()
                            if branch in configrepos[configproject]["branches"]:
                                configrepo = configrepos[configproject]["repo"]
                                workvalues.__setitem__("configrepo", configrepo)
                                workvalues.__setitem__("configproject", configproject)
                                ret.append(self.expand(variables, workvalues, drupal_domains_pattern, drupal_domain_pattern))
                    else:
                        ret.append(self.expand(variables, values, drupal_domains_pattern, drupal_domain_pattern))

            except AnsibleParserError:
                raise AnsibleError("could not access repository in lookup: %s" % term)

        if len(ret) > 1:
            return ret
        return [ret]

    def expand(self, variables, values, domain_key_pattern, domain_pattern):
        branches = variables.get('drupal_branches')
        domains = variables.get('drupal_domains')
        aliases = variables.get('drupal_domain_aliases')
        basepaths = variables.get('drupal_domain_basepaths')
        borgbackup = variables.get('drupal_borgbase')
        domain_key = self.replace(domain_key_pattern, values)
        if branches and domain_key in branches:
            values.__setitem__("branch", branches.get(domain_key))
        if aliases and domain_key in aliases:
            values.__setitem__("aliases", aliases.get(domain_key))
        if basepaths and values.get("configproject") in basepaths:
            values.__setitem__("basepath", basepaths.get(values.get("configproject")))
        if borgbackup:
            if domain_key in borgbackup:
                values.__setitem__("borg_enabled", 'yes')
                values.__setitem__("borg_host", borgbackup.get(domain_key).get('borg_host'))
                values.__setitem__("borg_repo", borgbackup.get(domain_key).get('borg_repo'))
            else:
                values.__setitem__("borg_enabled", 'no')
                values.__setitem__("borg_host", '')
                values.__setitem__("borg_repo", '')
        if domain_key in domains:
            domain = domains.get(domain_key)
        else:
            domain = self.replace(domain_pattern, values)
        values.__setitem__("domain", domain.replace('_', '-'))
        return self._expand(dict(variables.get('drupal_settings_template')), values)

    def _expand(self, itemplate, values):
        template = dict(itemplate)
        for key in template:
            if isinstance(template.get(key), string_types):
                if key == 'aliases':
                    template.__setitem__(key, self.aliases(values))
                else:
                    template.__setitem__(key, self.replace(template.get(key), values))
            elif isinstance(template.get(key), dict):
                template.__setitem__(key, self._expand(template.get(key), values))
            elif isinstance(template.get(key), list):
                expanded = []
                for item in template.get(key):
                    if isinstance(item, string_types):
                        expanded.append(self.replace(item, values))
                    elif isinstance(item, dict):
                        expanded.append(self._expand(item, values))
                template.__setitem__(key, expanded)
        return template

    @staticmethod
    def replace(string, values):
        for key in values:
            if key != 'aliases':
                string = string.replace('___%s___' % key, values.get(key))
        return string

    def aliases(self, values):
        aliases = [self.replace('___branch2url___.___configproject___.___project___.mobimo.lakedrops.com', values).replace('_', '-')]
        if 'aliases' in values:
            for alias in values['aliases']:
                aliases.append(alias.replace('_', '-'))
        return aliases
